# ZpCron - Cron-like scheduler for Zope
# Copyright (C) 2008  Jan Janech <janech@kios.sk>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import datetime
import calendar
import sys
import bisect

class CronTab(object):
    classes = {
        "@reboot":      'reboot',
        
        "@yearly":      ("0", "0", "1", "1", "*"),
        "@annually":    ("0", "0", "1", "1", "*"),
        "@monthly":     ("0", "0", "1", "*", "*"),
        "@weekly":      ("0", "0", "*", "*", "0"),
        "@daily":       ("0", "0", "*", "*", "*"),
        "@midnight":    ("0", "0", "*", "*", "*"),
        "@hourly":      ("0", "*", "*", "*", "*")
    }
    
    def __init__(self, file = None):
        self.reset()
        if file is not None:
            self.parseFile(file)
    
    ### <override this> ####
    def reset(self):
        self.__restart = True
        self.__lines = []
        self.__rebootlines = []
    
    def getRestart(self):
        r = self.__restart
        self.__restart = False
        return r
    
    def getLines(self):
        return self.__lines
    
    def getRebootLines(self):
        return self.__rebootlines
    
    def postparseCommand(self, command):
        return command
    ### </override this> ###
    
    def getMinMax(self, times, type):
#        if type == 2:
#            monthmax = (31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
#            month = times[3]
#            return 1, monthmax[month]
#        else:
        minmax = ((0, 59), (0, 23), (1, 29), (1, 12), (0, 7))
        return minmax[type]
    
    def normalizeTimes(self, times):
        ret = []
        for type, t1 in enumerate(times):
            min, max = self.getMinMax(ret, type)
            x = []
            for t in t1.split(','):
                t = t.split('/')
                if len(t) > 1:
                    interval = int(t[1])
                else:
                    interval = 1
                if t[0] == '*':
                    x.extend(range(min, max, interval))
                else:
                    t = t[0].split('-')
                    if len(t) > 1:
                        x.extend(range(int(t[0]), int(t[1])+1, interval))
                    else:
                        x.append(int(t[0]))
            x = list(set(x))
            x.sort()
            ret.append(x)
        tmp = ret[-1]
        ret[-1] = set()
        for i in tmp:
            ret[-1].add((i+6)%7)
        return ret
    
    def addLine(self, extime, cmd, now = None):
        if now is None:
            now = time.time()
        
        if extime == '@reboot':
            rebootlines = self.getRebootLines()
            rebootlines.append(cmd)
        else:
            lines = self.getLines()
            bisect.insort(lines, (self.calculateNextTime(extime, now = now), extime, cmd))
    
    def parseFile(self, file):
        variables = {}
        for line in file:
            line = line.strip()
            if not line or line[0] in ('#', ';'):
                continue
            if line[0].isalpha():
                var, val = line.split('=', 1)
                var, val = var.strip(), val.strip()
                if not val:
                    del variables[var]
                else:
                    if val.startswith('"') and val.endswith('"'):
                        val = val[1:-1]
                    variables[var] = val
            else:
                if line[0] == '@':
                    line = line.split(None, 1)
                    if line[0] != '@reboot':
                        line = self.normalizeTimes(self.classes[line[0]]), self.postparseCommand(line[1], **variables)
                    else:
                        line = line[0], self.postparseCommand(line[1], **variables)
                else:
                    line = line.split(None, 5)
                    line = self.normalizeTimes(line[:5]), self.postparseCommand(line[5], **variables)
                self.addLine(*line)
    
    def calculateNextTime(self, extime, now = None):
        if now is None:
            now = time.time()
        year, mon, day, hour, min, sec, wday, yday, isdst = time.localtime(now)
        MIN, HOUR, DAY, MON, WDAY = extime
        for i in MIN:
            if i > min:
                min = i
                break
        else:
            min = MIN[0]
            for i in HOUR:
                if i > hour:
                    hour = i
                    break
            else:
                hour = HOUR[0]
                while True:
                    for i in DAY:
                        if i > day:
                            day = i
                            break
                    else:
                        day = DAY[0]
                        for i in MON:
                            if i > mon:
                                mon = i
                                break
                        else:
                            mon = MON[0]
                    if calendar.weekday(year, mon, day) in WDAY:
                        break
        return time.mktime((year, mon, day, hour, min, 0, wday, -1, -1))
    
    def getFirstEvents(self, now = None):
        if now is None:
            now = time.time()
        
        lines = self.getLines()
        rebootlines = self.getRebootLines()
        
        ret = []
        if self.getRestart():
            ret.extend((-1, i) for i in rebootlines)
        ret.extend((-1, cmd) for runtime, extime, cmd in lines if runtime <= now)
        
        if ret:
            return ret
        if lines:
            return [(lines[0][0] - now, lines[0][2])]
        return []
    
    def refresh(self, now = None):
        lines = self.getLines()
        if not lines:
            return
        if now is None:
            now = time.time()
        
        while lines[0][0] <= now:
            runtime, extime, cmd = lines.pop(0)
            self.addLine(extime, cmd, now = now)
    
    def doit(self):
        while True:
            for waittime, cmd in self.getFirstEvents():
                if waittime <= 0:
                    yield waittime, cmd
            self.refresh()
            for waittime, cmd in self.getFirstEvents():
                if waittime > 0:
                    yield waittime, cmd

if __name__ == '__main__':
    import StringIO
    f = StringIO.StringIO("""
    @reboot     reboot
    @weekly     every week
    * * * * *   every minute
    */5 * * * *   every five minute
    44,*/3 * * * *   at fourty-four and every three minutes
    """)

    cron = CronTab(f)
    for waittime, cmd in cron.doit():
        if waittime <= 0:
            print 'CMD:  %r'%cmd
        else:
            print 'WAIT: %r'%waittime
            time.sleep(waittime+10)
